@@ -1,3 +1,5 @@
+{-# LANGUAGE LambdaCase #-}
+
 -- | Re-export of the @Csr@ module and generic graph search functions.
 --
 -- @since 1.1.0.0
@@ -58,7 +60,7 @@ import Data.Vector.Unboxed.Mutable qualified as VUM
 --
 -- @since 1.1.0.0
 {-# INLINE swapDupe #-}
-swapDupe :: (VU.Unbox (Int, Int, w)) => VU.Vector (Int, Int, w) -> VU.Vector (Int, Int, w)
+swapDupe :: (VU.Unbox w) => VU.Vector (Int, Int, w) -> VU.Vector (Int, Int, w)
 swapDupe = VU.concatMap (\(!u, !v, !w) -> VU.fromListN 2 [(u, v, w), (v, u, w)])
 
 -- | \(O(n)\) Converts non-directed edges into directional edges. This is a convenient function for
@@ -86,7 +88,7 @@ swapDupe = VU.concatMap (\(!u, !v, !w) -> VU.fromListN 2 [(u, v, w), (v, u, w)])
 --
 -- @since 1.1.0.0
 {-# INLINE swapDupe' #-}
-swapDupe' :: (VU.Unbox (Int, Int)) => VU.Vector (Int, Int) -> VU.Vector (Int, Int)
+swapDupe' :: VU.Vector (Int, Int) -> VU.Vector (Int, Int)
 swapDupe' = VU.concatMap (\(!u, !v) -> VU.fromListN 2 [(u, v), (v, u)])
 
 -- | \(O(n + m)\) Returns the strongly connected components.
@@ -121,7 +123,7 @@ scc = ACISCC.sccCsr
 -- [1,2,4,0,3]
 --
 -- @since 1.1.0.0
-{-# INLINABLE topSort #-}
+{-# INLINEABLE topSort #-}
 topSort :: Int -> (Int -> VU.Vector Int) -> VU.Vector Int
 topSort n gr = runST $ do
   inDeg <- VUM.replicate n (0 :: Int)
@@ -170,7 +172,7 @@ topSort n gr = runST $ do
 -- [[3,2],[0,3,1]]
 --
 -- @since 1.1.1.0
-{-# INLINABLE blockCut #-}
+{-# INLINEABLE blockCut #-}
 blockCut :: Int -> (Int -> VU.Vector Int) -> Csr ()
 blockCut n gr = runST $ do
   low <- VUM.replicate n (0 :: Int)
