@@ -1,3 +1,4 @@
+{-# LANGUAGE LambdaCase #-}
 {-# LANGUAGE RecordWildCards #-}
 
 -- | A fast, mutable multiset for `Int` keys backed by a @HashMap@.  Most operations are performed
@@ -45,7 +46,7 @@
 --
 -- >>> MS.inc ms 11
 -- >>> MS.sub ms 11 2
--- *** Exception: AtCoder.Extra.Multiset.sub: the count of `11` is becoming a negative value: `-1`
+-- *** Exception: AtCoder.Extra.Multiset.subST: the count of `11` is becoming a negative value: `-1`
 -- ...
 --
 -- Decrementing a non-existing key does nothing and does not throw an exception:
@@ -99,7 +100,8 @@ where
 
 import AtCoder.Extra.HashMap qualified as HM
 import Control.Monad (when)
-import Control.Monad.Primitive (PrimMonad, PrimState)
+import Control.Monad.Primitive (PrimMonad, PrimState, stToPrim)
+import Control.Monad.ST (ST)
 import Data.Functor ((<&>))
 import Data.Vector.Generic.Mutable qualified as VGM
 import Data.Vector.Unboxed qualified as VU
@@ -120,10 +122,7 @@ data MultiSet s = MultiSet
 -- @since 1.1.0.0
 {-# INLINE new #-}
 new :: (PrimMonad m) => Int -> m (MultiSet (PrimState m))
-new n = do
-  mapMS <- HM.new n
-  cntMS <- VUM.replicate 1 0
-  pure $ MultiSet {..}
+new n = stToPrim $ newST n
 
 -- | \(O(1)\) Returns the maximum number of distinct keys that can be inserted into the internal
 -- hash map.
@@ -146,27 +145,21 @@ size MultiSet {..} = do
 -- @since 1.1.0.0
 {-# INLINE lookup #-}
 lookup :: (PrimMonad m) => MultiSet (PrimState m) -> Int -> m (Maybe Int)
-lookup MultiSet {..} k = do
-  HM.lookup mapMS k <&> \case
-    Just i | i > 0 -> Just i
-    _ -> Nothing
+lookup ms k = stToPrim $ lookupST ms k
 
 -- | \(O(1)\) Tests whether \(k\) is in the set.
 --
 -- @since 1.1.0.0
 {-# INLINE member #-}
 member :: (PrimMonad m) => MultiSet (PrimState m) -> Int -> m Bool
-member MultiSet {..} k = do
-  HM.lookup mapMS k <&> \case
-    Just i -> i > 0
-    _ -> False
+member ms k = stToPrim $ memberST ms k
 
 -- | \(O(1)\) Tests whether \(k\) is not in the set.
 --
 -- @since 1.1.0.0
 {-# INLINE notMember #-}
 notMember :: (PrimMonad m) => MultiSet (PrimState m) -> Int -> m Bool
-notMember ms k = not <$> member ms k
+notMember ms k = stToPrim $ not <$> memberST ms k
 
 -- | \(O(1)\) Increments the count of a key.
 --
@@ -188,18 +181,7 @@ dec ms k = sub ms k 1
 -- @since 1.1.0.0
 {-# INLINE add #-}
 add :: (HasCallStack, PrimMonad m) => MultiSet (PrimState m) -> Int -> Int -> m ()
-add ms@MultiSet {..} k v = case compare v 0 of
-  LT -> sub ms k (-v)
-  EQ -> pure ()
-  GT -> do
-    HM.lookup mapMS k >>= \case
-      Just n ->  do
-        HM.insert mapMS k $ n + v
-        when (n <= 0) $ do
-          VGM.unsafeModify cntMS (+ 1) 0
-      Nothing -> do
-        HM.insert mapMS k v
-        VGM.unsafeModify cntMS (+ 1) 0
+add ms k v = stToPrim $ addST ms k v
 
 -- | \(O(1)\) Decrements the count of a key \(k\) by \(c\). If \(c\) is negative, it falls back to
 -- `add`.
@@ -207,35 +189,14 @@ add ms@MultiSet {..} k v = case compare v 0 of
 -- @since 1.1.0.0
 {-# INLINE sub #-}
 sub :: (HasCallStack, PrimMonad m) => MultiSet (PrimState m) -> Int -> Int -> m ()
-sub ms@MultiSet {..} k v = case compare v 0 of
-  LT -> add ms k (-v)
-  EQ -> pure ()
-  GT -> do
-    HM.lookup mapMS k >>= \case
-      Just 0 -> pure () -- ignored
-      Just n -> case compare n v of
-        GT -> do
-          HM.insert mapMS k (n - v)
-        EQ -> do
-          HM.insert mapMS k 0
-          VGM.unsafeModify cntMS (subtract 1) 0
-        LT -> error $ "AtCoder.Extra.Multiset.sub: the count of `" ++ show k ++ "` is becoming a negative value: `" ++ show (n - v) ++ "`"
-      _ -> pure ()
+sub ms k v = stToPrim $ subST ms k v
 
 -- | \(O(1)\) Inserts a key-count pair into the set. `MultiSet` is actually a count map.
 --
 -- @since 1.1.0.0
 {-# INLINE insert #-}
 insert :: (HasCallStack, PrimMonad m) => MultiSet (PrimState m) -> Int -> Int -> m ()
-insert MultiSet {..} k v
-  | v <= 0 = error $ "AtCoder.Extra.Multiset.insert: new count must be positive`" ++ show k ++ "`: `" ++ show v ++ "`"
-  | otherwise = do
-      HM.lookup mapMS k >>= \case
-        Just n | n > 0 -> do
-          HM.insert mapMS k v
-        _ -> do
-          HM.insert mapMS k v
-          VGM.unsafeModify cntMS (+ 1) 0
+insert ms k v = stToPrim $ insertST ms k v
 
 -- | \(O(1)\) Deletes a key. Note that it does not undo its insertion and does not increase the
 -- number of distinct keys that can be inserted into the internal hash map.
@@ -243,12 +204,7 @@ insert MultiSet {..} k v
 -- @since 1.1.0.0
 {-# INLINE delete #-}
 delete :: (HasCallStack, PrimMonad m) => MultiSet (PrimState m) -> Int -> m ()
-delete MultiSet {..} k = do
-  HM.lookup mapMS k >>= \case
-    Just i | i > 0 -> do
-      HM.insert mapMS k 0
-      VGM.unsafeModify cntMS (subtract 1) 0
-    _ -> pure ()
+delete ms k = stToPrim $ deleteST ms k
 
 -- | \(O(n)\) Enumerates the keys in the set.
 --
@@ -291,3 +247,81 @@ unsafeElems = (VU.filter (> 0) <$>) . HM.unsafeElems . mapMS
 {-# INLINE unsafeAssocs #-}
 unsafeAssocs :: (PrimMonad m) => MultiSet (PrimState m) -> m (VU.Vector (Int, Int))
 unsafeAssocs = (VU.filter (\(!_, !n) -> n > 0) <$>) . HM.unsafeAssocs . mapMS
+
+-- -------------------------------------------------------------------------------
+-- Internal
+-- -------------------------------------------------------------------------------
+
+{-# INLINEABLE newST #-}
+newST :: Int -> ST s (MultiSet s)
+newST n = do
+  mapMS <- HM.new n
+  cntMS <- VUM.replicate 1 0
+  pure $ MultiSet {..}
+
+{-# INLINEABLE lookupST #-}
+lookupST :: MultiSet s -> Int -> ST s (Maybe Int)
+lookupST MultiSet {..} k = do
+  HM.lookup mapMS k <&> \case
+    Just i | i > 0 -> Just i
+    _ -> Nothing
+
+{-# INLINEABLE memberST #-}
+memberST :: MultiSet s -> Int -> ST s Bool
+memberST MultiSet {..} k = do
+  HM.lookup mapMS k <&> \case
+    Just i -> i > 0
+    _ -> False
+
+{-# INLINEABLE addST #-}
+addST :: (HasCallStack) => MultiSet s -> Int -> Int -> ST s ()
+addST ms@MultiSet {..} k v = case compare v 0 of
+  LT -> subST ms k (-v)
+  EQ -> pure ()
+  GT -> do
+    HM.lookup mapMS k >>= \case
+      Just n -> do
+        HM.insert mapMS k $ n + v
+        when (n <= 0) $ do
+          VGM.unsafeModify cntMS (+ 1) 0
+      Nothing -> do
+        HM.insert mapMS k v
+        VGM.unsafeModify cntMS (+ 1) 0
+
+{-# INLINEABLE subST #-}
+subST :: (HasCallStack) => MultiSet s -> Int -> Int -> ST s ()
+subST ms@MultiSet {..} k v = case compare v 0 of
+  LT -> addST ms k (-v)
+  EQ -> pure ()
+  GT -> do
+    HM.lookup mapMS k >>= \case
+      Just 0 -> pure () -- ignored
+      Just n -> case compare n v of
+        GT -> do
+          HM.insert mapMS k (n - v)
+        EQ -> do
+          HM.insert mapMS k 0
+          VGM.unsafeModify cntMS (subtract 1) 0
+        LT -> error $ "AtCoder.Extra.Multiset.subST: the count of `" ++ show k ++ "` is becoming a negative value: `" ++ show (n - v) ++ "`"
+      _ -> pure ()
+
+{-# INLINEABLE insertST #-}
+insertST :: (HasCallStack) => MultiSet s -> Int -> Int -> ST s ()
+insertST MultiSet {..} k v
+  | v <= 0 = error $ "AtCoder.Extra.Multiset.insertST: new count must be positive`" ++ show k ++ "`: `" ++ show v ++ "`"
+  | otherwise = do
+      HM.lookup mapMS k >>= \case
+        Just n | n > 0 -> do
+          HM.insert mapMS k v
+        _ -> do
+          HM.insert mapMS k v
+          VGM.unsafeModify cntMS (+ 1) 0
+
+{-# INLINEABLE deleteST #-}
+deleteST :: (HasCallStack) => MultiSet s -> Int -> ST s ()
+deleteST MultiSet {..} k = do
+  HM.lookup mapMS k >>= \case
+    Just i | i > 0 -> do
+      HM.insert mapMS k 0
+      VGM.unsafeModify cntMS (subtract 1) 0
+    _ -> pure ()
