@@ -3,13 +3,16 @@
 
 -- | Dynamic sequence of monoid values with monoid actions on them through the `SegAct` instance.
 --
+-- ==== Performance
+-- This module is __slow__ as an ordinary dynamic sequence. Consider using another module if you
+-- don't need monoid products.
+--
 -- ==== __Example__
 --
 -- Create a `Seq` storage of length \(10\):
 --
 -- >>> import AtCoder.Extra.Monoid.RangeAdd qualified as RangeAdd
 -- >>> import AtCoder.Extra.Seq qualified as Seq
--- >>> import AtCoder.LazySegTree (SegAct (..))
 -- >>> import Data.Semigroup (Sum (..))
 -- >>> import Data.Vector.Unboxed qualified as VU
 -- >>> seq <- Seq.new @_ @(RangeAdd.RangeAdd (Sum Int)) @(Sum Int) 10
@@ -96,11 +99,17 @@
 module AtCoder.Extra.Seq
   ( -- * Seq
     Seq.Seq (..),
+
+    -- * Handle (re-exports)
     Handle (..),
+    -- TODO: hide
     newHandle,
     nullHandle,
     invalidateHandle,
 
+    -- * Re-exports
+    SegAct (..),
+
     -- * Constructors
     new,
     reset,
@@ -108,6 +117,10 @@ module AtCoder.Extra.Seq
     newNode,
     newSeq,
 
+    -- * Metadata
+    capacity,
+    length,
+
     -- * Merge/split
     merge,
     merge3,
@@ -160,6 +173,7 @@ module AtCoder.Extra.Seq
   )
 where
 
+import AtCoder.Extra.Pool (Handle (..), invalidateHandle, newHandle, nullHandle)
 import AtCoder.Extra.Pool qualified as P
 import AtCoder.Extra.Seq.Raw (Seq (..))
 import AtCoder.Extra.Seq.Raw qualified as Seq
@@ -167,18 +181,8 @@ import AtCoder.LazySegTree (SegAct (..))
 import Control.Monad.Primitive (PrimMonad, PrimState, stToPrim)
 import Data.Vector.Generic.Mutable qualified as VGM
 import Data.Vector.Unboxed qualified as VU
-import Data.Vector.Unboxed.Mutable qualified as VUM
 import GHC.Stack (HasCallStack)
-import Prelude hiding (read, reverse, seq)
-
--- | `Handle` for a sequence in `Seq`. It internally stores the root node and updates it
--- following splaying operations, as `Seq` utilizes a splay tree structure.
---
--- @since 1.2.0.0
-newtype Handle s = Handle
-  { -- | @since 1.2.0.0
-    unHandle :: VUM.MVector s P.Index
-  }
+import Prelude hiding (length, read, reverse, seq)
 
 -- | \(O(n)\) Creates a new `Seq` of length \(n\).
 --
@@ -187,29 +191,6 @@ newtype Handle s = Handle
 new :: (PrimMonad m, Monoid f, VU.Unbox f, Monoid a, VU.Unbox a) => Int -> m (Seq (PrimState m) f a)
 new n = stToPrim $ Seq.newST n
 
--- | \(O(1)\) Creates a new sequence `Handle` from a root node index.
---
--- @since 1.2.0.0
-{-# INLINE newHandle #-}
-newHandle :: (PrimMonad m) => P.Index -> m (Handle (PrimState m))
-newHandle x = stToPrim $ Handle <$> VUM.replicate 1 x
-
--- | \(O(1)\) Returns whether the sequence is empty.
---
--- @since 1.2.0.0
-{-# INLINE nullHandle #-}
-nullHandle :: (PrimMonad m) => Handle (PrimState m) -> m Bool
-nullHandle (Handle h) = stToPrim $ do
-  P.nullIndex <$> VGM.unsafeRead h 0
-
--- | \(O(1)\) Invalidates a sequence handle. Note that it does not change or `free` the sequence.
---
--- @since 1.2.0.0
-{-# INLINE invalidateHandle #-}
-invalidateHandle :: (PrimMonad m) => Handle (PrimState m) -> m ()
-invalidateHandle (Handle h) = stToPrim $ do
-  VGM.unsafeWrite h 0 P.undefIndex
-
 -- | \(O(1)\) Clears the sequence storage. All the handles must not be used again.
 --
 -- @since 1.2.0.0
@@ -221,14 +202,14 @@ reset seq = stToPrim $ Seq.resetST seq
 --
 -- @since 1.2.0.0
 {-# INLINE newNode #-}
-newNode :: (PrimMonad m, Monoid f, VU.Unbox f, VU.Unbox a) => Seq (PrimState m) f a -> a -> m (Handle (PrimState m))
+newNode :: (HasCallStack, PrimMonad m, Monoid f, VU.Unbox f, VU.Unbox a) => Seq (PrimState m) f a -> a -> m (Handle (PrimState m))
 newNode seq x = stToPrim $ newHandle =<< Seq.newNodeST seq x
 
 -- | \(O(n)\) Allocates a new sequence.
 --
 -- @since 1.2.0.0
 {-# INLINE newSeq #-}
-newSeq :: (PrimMonad m, Monoid f, VU.Unbox f, Monoid a, VU.Unbox a) => Seq (PrimState m) f a -> VU.Vector a -> m (Handle (PrimState m))
+newSeq :: (HasCallStack, PrimMonad m, Monoid f, VU.Unbox f, Monoid a, VU.Unbox a) => Seq (PrimState m) f a -> VU.Vector a -> m (Handle (PrimState m))
 newSeq seq !xs = stToPrim $ newHandle =<< Seq.newSeqST seq xs
 
 -- | \(O(n)\) Frees a sequence and invalidates the handle.
@@ -241,6 +222,26 @@ free seq (Handle handle) = stToPrim $ do
   Seq.freeSubtreeST seq c0
   VGM.write handle 0 P.undefIndex
 
+-- -------------------------------------------------------------------------------------------------
+-- Meta
+-- -------------------------------------------------------------------------------------------------
+
+-- | \(O(1)\) Returns the capacity of the sequence storage.
+--
+-- @since 1.2.1.0
+{-# INLINE capacity #-}
+capacity :: Seq s f a -> Int
+capacity = Seq.capacity
+
+-- | \(O(1)\) Returns the length of the sequence.
+--
+-- @since 1.2.1.0
+{-# INLINE length #-}
+length :: (PrimMonad m) => Seq (PrimState m) f a -> Handle (PrimState m) -> m Int
+length seq (Handle handle) = stToPrim $ do
+  i <- VGM.unsafeRead handle 0
+  Seq.lengthST seq i
+
 -- -------------------------------------------------------------------------------------------------
 -- Merge/split
 -- -------------------------------------------------------------------------------------------------
@@ -437,7 +438,7 @@ exchange seq (Handle hRoot) k v = stToPrim $ do
 --
 -- @since 1.2.0.0
 {-# INLINE prod #-}
-prod :: (HasCallStack, Show a, PrimMonad m, SegAct f a, Eq f, Monoid f, VU.Unbox f, Monoid a, VU.Unbox a) => Seq (PrimState m) f a -> Handle (PrimState m) -> Int -> Int -> m a
+prod :: (HasCallStack, PrimMonad m, SegAct f a, Eq f, Monoid f, VU.Unbox f, Monoid a, VU.Unbox a) => Seq (PrimState m) f a -> Handle (PrimState m) -> Int -> Int -> m a
 prod seq (Handle hRoot) l r = stToPrim $ do
   root <- VGM.unsafeRead hRoot 0
   (!v, !root') <- Seq.prodST seq root l r
@@ -445,19 +446,22 @@ prod seq (Handle hRoot) l r = stToPrim $ do
   pure v
 
 -- | Amortized \(O(\log n)\). Returns the monoid product in an interval \([l, r)\). Returns
--- `Nothing` if an invalid interval is given.
+-- `Nothing` if the interval is invalid.
 --
 -- @since 1.2.0.0
-{-# INLINE prodMaybe #-}
+{-# INLINEABLE prodMaybe #-}
 prodMaybe :: (HasCallStack, PrimMonad m, SegAct f a, Eq f, Monoid f, VU.Unbox f, Monoid a, VU.Unbox a) => Seq (PrimState m) f a -> Handle (PrimState m) -> Int -> Int -> m (Maybe a)
 prodMaybe seq (Handle handle) l r = stToPrim $ do
   root <- VGM.unsafeRead handle 0
-  res <- Seq.prodMaybeST seq root l r
-  case res of
-    Just (!v, !root') -> do
-      VGM.unsafeWrite handle 0 root'
-      pure $ Just v
-    Nothing -> pure Nothing
+  if P.nullIndex root
+    then pure $ Just mempty
+    else do
+      res <- Seq.prodMaybeST seq root l r
+      case res of
+        Just (!v, !root') -> do
+          VGM.unsafeWrite handle 0 root'
+          pure $ Just v
+        Nothing -> pure Nothing
 
 -- | Amortized \(O(\log n)\). Returns the monoid product of the whole sequence.
 --
@@ -577,9 +581,6 @@ detach seq (Handle hRoot) i = stToPrim $ do
 
 -- | Amortized \(O(\log n)\).
 --
--- ==== Constraints
--- - The sequence must be non-empty.
---
 -- @since 1.2.0.0
 {-# INLINE ilowerBound #-}
 ilowerBound ::
@@ -600,9 +601,6 @@ ilowerBound seq (Handle root0) f = stToPrim $ do
 
 -- | Amortized \(O(\log n)\).
 --
--- ==== Constraints
--- - The sequence must be non-empty.
---
 -- @since 1.2.0.0
 {-# INLINE ilowerBoundM #-}
 ilowerBoundM ::
@@ -623,9 +621,6 @@ ilowerBoundM seq (Handle root0) f = do
 
 -- | Amortized \(O(\log n)\).
 --
--- ==== Constraints
--- - The sequence must be non-empty.
---
 -- @since 1.2.0.0
 {-# INLINE ilowerBoundProd #-}
 ilowerBoundProd ::
@@ -646,9 +641,6 @@ ilowerBoundProd seq (Handle root0) f = stToPrim $ do
 
 -- | Amortized \(O(\log n)\).
 --
--- ==== Constraints
--- - The sequence must be non-empty.
---
 -- @since 1.2.0.0
 {-# INLINE ilowerBoundProdM #-}
 ilowerBoundProdM ::
@@ -670,9 +662,6 @@ ilowerBoundProdM seq (Handle root0) f = do
 -- | Amortized \(O(\log n)\). Splits a sequence into two with the user predicate and returns the
 -- right sequence handle.
 --
--- ==== Constraints
--- - The sequence must be non-empty.
---
 -- @since 1.2.0.0
 {-# INLINE isplitMaxRight #-}
 isplitMaxRight ::
@@ -695,9 +684,6 @@ isplitMaxRight seq (Handle root0) f = stToPrim $ do
 -- | Amortized \(O(\log n)\). Splits a sequence into two with the user predicate and returns the
 -- right sequence handle.
 --
--- ==== Constraints
--- - The sequence must be non-empty.
---
 -- @since 1.2.0.0
 {-# INLINEABLE isplitMaxRightM #-}
 isplitMaxRightM ::
@@ -720,9 +706,6 @@ isplitMaxRightM seq (Handle root0) f = do
 -- | Amortized \(O(\log n)\). Splits a sequence into two with the user predicate and returns the
 -- right sequence handle.
 --
--- ==== Constraints
--- - The sequence must be non-empty.
---
 -- @since 1.2.0.0
 {-# INLINE isplitMaxRightProd #-}
 isplitMaxRightProd ::
@@ -745,9 +728,6 @@ isplitMaxRightProd seq (Handle root0) f = stToPrim $ do
 -- | Amortized \(O(\log n)\). Splits a sequence into two with the user predicate and returns the
 -- right sequence handle.
 --
--- ==== Constraints
--- - The sequence must be non-empty.
---
 -- @since 1.2.0.0
 {-# INLINEABLE isplitMaxRightProdM #-}
 isplitMaxRightProdM ::
