@@ -1,6 +1,7 @@
 {-# LANGUAGE DerivingVia #-}
 {-# LANGUAGE RecordWildCards #-}
 {-# LANGUAGE TypeFamilies #-}
+{-# OPTIONS_HADDOCK hide #-}
 
 -- | Base module for implementing dynamic sequences. It internaly uses a splay tree and user has to
 -- track the root node change.
@@ -18,6 +19,10 @@ module AtCoder.Extra.Seq.Raw
     freeNodeST,
     freeSubtreeST,
 
+    -- * Metadata
+    capacity,
+    lengthST,
+
     -- * Merge/split
     mergeST,
     merge3ST,
@@ -78,6 +83,18 @@ module AtCoder.Extra.Seq.Raw
 
     -- * Conversions
     freezeST,
+
+    -- * Internals
+
+    -- | These functions are exported primarily for @Map@ implementations.
+    splitMaxRightWithST,
+    maxRightWithST,
+    updateNodeST,
+    writeNodeST,
+    modifyNodeST,
+    exchangeNodeST,
+    propNodeST,
+    applyNodeST,
   )
 where
 
@@ -173,7 +190,7 @@ resetST Seq {poolSeq} = stToPrim $ P.clear poolSeq
 --
 -- @since 1.2.0.0
 {-# INLINEABLE newNodeST #-}
-newNodeST :: (Monoid f, VU.Unbox f, VU.Unbox a) => Seq s f a -> a -> ST s P.Index
+newNodeST :: (HasCallStack, Monoid f, VU.Unbox f, VU.Unbox a) => Seq s f a -> a -> ST s P.Index
 newNodeST Seq {..} x = do
   i <- P.alloc poolSeq ()
   VGM.write lSeq (coerce i) P.undefIndex
@@ -190,7 +207,7 @@ newNodeST Seq {..} x = do
 --
 -- @since 1.2.0.0
 {-# INLINEABLE newSeqST #-}
-newSeqST :: (Monoid f, VU.Unbox f, Monoid a, VU.Unbox a) => Seq s f a -> VU.Vector a -> ST s P.Index
+newSeqST :: (HasCallStack, Monoid f, VU.Unbox f, Monoid a, VU.Unbox a) => Seq s f a -> VU.Vector a -> ST s P.Index
 newSeqST seq@Seq {..} !xs = do
   -- [l, r)
   let inner l r
@@ -199,8 +216,8 @@ newSeqST seq@Seq {..} !xs = do
         | otherwise = do
             let !m = (l + r) `div` 2
             rootL <- inner l m
-            rootR <- inner (m + 1) r
             root <- newNodeST seq (xs VG.! m)
+            rootR <- inner (m + 1) r
             unless (P.nullIndex rootL) $ do
               VGM.write lSeq (coerce root) rootL
               VGM.write pSeq (coerce rootL) root
@@ -234,6 +251,26 @@ freeSubtreeST Seq {lSeq, rSeq, poolSeq} c0
       inner c0
       P.free poolSeq c0
 
+-- -------------------------------------------------------------------------------------------------
+-- Metadata
+-- -------------------------------------------------------------------------------------------------
+
+-- | \(O(1)\) Returns the capacity of the sequence storage.
+--
+-- @since 1.2.1.0
+{-# INLINE capacity #-}
+capacity :: Seq s f a -> Int
+capacity = nSeq
+
+-- | \(O(1)\) Returns the length of a sequence or a subtree.
+--
+-- @since 1.2.1.0
+{-# INLINE lengthST #-}
+lengthST :: Seq s f a -> P.Index -> ST s Int
+lengthST Seq {..} i
+  | P.nullIndex i = pure 0
+  | otherwise = VGM.read sSeq (coerce i)
+
 -- -------------------------------------------------------------------------------------------------
 -- Merge/split
 -- -------------------------------------------------------------------------------------------------
@@ -250,9 +287,9 @@ assertRootOrNullST :: (HasCallStack) => Seq s f a -> P.Index -> ST s ()
 assertRootOrNullST Seq {pSeq} i
   | P.nullIndex i = pure ()
   | otherwise = do
-    p <- VGM.read pSeq (coerce i)
-    let !_ = ACIA.runtimeAssert (P.nullIndex p) $ "AtCoder.Extra.Seq.Raw.assertRootOrNullST: not a root (node `" ++ show i ++ "`, parent `" ++ show p ++ "`)"
-    pure ()
+      p <- VGM.read pSeq (coerce i)
+      let !_ = ACIA.runtimeAssert (P.nullIndex p) $ "AtCoder.Extra.Seq.Raw.assertRootOrNullST: not a root (node `" ++ show i ++ "`, parent `" ++ show p ++ "`)"
+      pure ()
 
 -- | Amortized \(O(\log n)\). Merges two sequences \(l, r\) into one in the given order, ignoring
 -- empty sequences.
@@ -267,12 +304,8 @@ mergeST seq@Seq {pSeq, lSeq} lRoot rRoot
   | P.nullIndex lRoot = pure rRoot
   | P.nullIndex rRoot = pure lRoot
   | otherwise = do
-      do
-        -- TODO: delete
-        lp <- VGM.read pSeq (coerce lRoot)
-        rp <- VGM.read pSeq (coerce rRoot)
-        let !_ = ACIA.runtimeAssert (lp == rp) "AtCoder.Extra.Seq.Raw.mergeST: given non-root node"
-        pure ()
+      assertRootST seq lRoot
+      assertRootST seq rRoot
       rRoot' <- splayKthST seq rRoot 0
       VGM.write lSeq (coerce rRoot') lRoot
       VGM.write pSeq (coerce lRoot) rRoot'
@@ -372,8 +405,8 @@ splitLrST seq@Seq {..} root = do
   s <- do
     rootL <- VGM.read lSeq (coerce root)
     if P.nullIndex rootL
-      then VGM.read sSeq (coerce rootL)
-      else pure 0
+      then pure 0
+      else VGM.read sSeq (coerce rootL)
   split3ST seq root s (s + 1)
 
 -- | Amortized \(O(\log n)\). Captures the root of a subtree of \([l, r)\). Splay the new root after
@@ -445,13 +478,13 @@ readMaybeST :: (HasCallStack, SegAct f a, Eq f, Monoid f, VU.Unbox f, Monoid a,
 readMaybeST seq@Seq {..} root k
   | P.nullIndex root = pure Nothing
   | otherwise = do
-    assertRootST seq root
-    s <- VGM.read sSeq (coerce root)
-    if 0 <= k && k < s
-      then do
-        root' <- splayKthST seq root k
-        Just . (,root') <$> VGM.read vSeq (coerce root')
-      else pure Nothing
+      assertRootST seq root
+      s <- VGM.read sSeq (coerce root)
+      if 0 <= k && k < s
+        then do
+          root' <- splayKthST seq root k
+          Just . (,root') <$> VGM.read vSeq (coerce root')
+        else pure Nothing
 
 -- | Amortized \(O(\log n)\). Writes to the \(k\)-th node's monoid value.
 --
@@ -523,13 +556,13 @@ prodMaybeST :: (HasCallStack, SegAct f a, Eq f, Monoid f, VU.Unbox f, Monoid a,
 prodMaybeST seq@Seq {sSeq} root l r
   | P.nullIndex root = pure Nothing
   | otherwise = do
-    s <- VGM.read sSeq (coerce root)
-    if not (ACIA.testInterval l r s)
-      then pure Nothing
-      else
-        if l == r
-          then pure $ Just (mempty, root)
-          else Just <$> unsafeProdST seq root l r
+      s <- VGM.read sSeq (coerce root)
+      if not (ACIA.testInterval l r s)
+        then pure Nothing
+        else
+          if l == r
+            then pure $ Just (mempty, root)
+            else Just <$> unsafeProdST seq root l r
 
 -- | Amortized \(O(\log n)\).
 --
@@ -564,13 +597,12 @@ prodAllST seq@Seq {..} root = do
 --
 -- ==== Constraints
 -- - \(0 \le l \le r \le n\)
--- - The root must point to a non-empty sequence.
 --
 -- @since 1.2.0.0
 {-# INLINEABLE applyInST #-}
 applyInST :: (HasCallStack, SegAct f a, Eq f, Monoid f, VU.Unbox f, Monoid a, VU.Unbox a) => Seq s f a -> P.Index -> Int -> Int -> f -> ST s P.Index
 applyInST seq@Seq {..} root l r act = do
-  assertRootST seq root
+  -- assertRootOrNullST seq root
   s <- if P.nullIndex root then pure 0 else VGM.read sSeq (coerce root)
   let !_ = ACIA.checkInterval "AtCoder.Extra.Seq.applyInST" l r s
   if l == r
@@ -589,9 +621,9 @@ applyToRootST :: (HasCallStack, SegAct f a, Eq f, Monoid f, VU.Unbox f, Monoid a
 applyToRootST seq@Seq {..} root act
   | P.nullIndex root = pure ()
   | otherwise = do
-    rootP <- VGM.read pSeq (coerce root)
-    when (P.nullIndex rootP) $ do
-      applyNodeST seq root act
+      rootP <- VGM.read pSeq (coerce root)
+      when (P.nullIndex rootP) $ do
+        applyNodeST seq root act
 
 -- | Amortized \(O(\log n)\). Reverses the sequence in \([l, r)\).
 --
@@ -605,17 +637,17 @@ reverseST :: (HasCallStack, SegAct f a, Eq f, Monoid f, VU.Unbox f, Monoid a, VU
 reverseST seq@Seq {sSeq} root0 l r
   | P.nullIndex root0 = pure P.undefIndex
   | otherwise = do
-    s <- VGM.read sSeq (coerce root0)
-    if not (ACIA.testInterval l r s)
-      then pure root0
-      else
-        if l == r
-          then pure root0
-          else do
-            root' <- sliceST seq root0 l r
-            reverseNodeST seq root'
-            splayST seq root' True
-            pure root'
+      s <- VGM.read sSeq (coerce root0)
+      if not (ACIA.testInterval l r s)
+        then pure root0
+        else
+          if l == r
+            then pure root0
+            else do
+              root' <- sliceST seq root0 l r
+              reverseNodeST seq root'
+              splayST seq root' True
+              pure root'
 
 -- | Amortized \(O(\log n)\). Inserts a new node at \(k\) with initial monoid value \(v\). This
 -- functions for an empty index.
@@ -745,6 +777,7 @@ rotateST Seq {..} !i = do
 {-# INLINEABLE splayST #-}
 splayST :: (HasCallStack, SegAct f a, Eq f, Monoid f, VU.Unbox f, Monoid a, VU.Unbox a) => Seq s f a -> P.Index -> Bool -> ST s ()
 splayST seq@Seq {..} i doneParentProp = do
+  -- we have no chance to call `splayST` with `doneParentProp == False` though
   if doneParentProp
     then propNodeST seq i
     else propNodeFromRootST seq i
@@ -812,9 +845,6 @@ splayKthST seq@Seq {..} root0 k0 = do
 
 -- | Amortized \(O(\log n)\).
 --
--- ==== Constraints
--- - The node must be a root.
---
 -- @since 1.2.0.0
 {-# INLINE ilowerBoundST #-}
 ilowerBoundST ::
@@ -828,15 +858,14 @@ ilowerBoundST ::
   -- | (r, root)
   ST s (Int, P.Index)
 ilowerBoundST seq root f = stToPrim $ do
-  (!r, !_, !root') <- imaxRightST seq root f
-  splayST seq root' True
-  pure (r, root')
+  if P.nullIndex root
+    then pure (0, P.undefIndex)
+    else do
+      (!r, !_, !root') <- imaxRightST seq root f
+      pure (r, root')
 
 -- | Amortized \(O(\log n)\).
 --
--- ==== Constraints
--- - The node must be a root.
---
 -- @since 1.2.0.0
 {-# INLINE ilowerBoundM #-}
 ilowerBoundM ::
@@ -850,9 +879,11 @@ ilowerBoundM ::
   -- | (r, root)
   m (Int, P.Index)
 ilowerBoundM seq root f = do
-  (!r, !_, !root') <- imaxRightM seq root f
-  stToPrim $ splayST seq root' True
-  pure (r, root')
+  if P.nullIndex root
+    then pure (0, P.undefIndex)
+    else do
+      (!r, !_, !root') <- imaxRightM seq root f
+      pure (r, root')
 
 -- | Amortized \(O(\log n)\).
 --
@@ -872,8 +903,11 @@ ilowerBoundProdST ::
   -- | (r, root)
   ST s (Int, P.Index)
 ilowerBoundProdST seq root f = do
-  (!r, !_, !root') <- imaxRightProdST seq root f
-  pure (r, root')
+  if P.nullIndex root
+    then pure (0, P.undefIndex)
+    else do
+      (!r, !_, !root') <- imaxRightProdST seq root f
+      pure (r, root')
 
 -- | Amortized \(O(\log n)\).
 --
@@ -893,15 +927,15 @@ ilowerBoundProdM ::
   -- | (r, root)
   m (Int, P.Index)
 ilowerBoundProdM seq root f = do
-  (!r, !_, !root') <- imaxRightProdM seq root f
-  pure (r, root')
+  if P.nullIndex root
+    then pure (0, P.undefIndex)
+    else do
+      (!r, !_, !root') <- imaxRightProdM seq root f
+      pure (r, root')
 
 -- | Amortized \(O(\log n)\). Given a monotonious sequence, returns the rightmost node \(v_k\)
 -- where \(f(v)\) holds for every \([0, i) (0 \le i \lt k)\).
 --
--- ==== Constraints
--- - The node must be a root.
---
 -- @since 1.2.0.0
 {-# INLINE isplitMaxRightST #-}
 isplitMaxRightST ::
@@ -919,9 +953,6 @@ isplitMaxRightST seq root f = stToPrim $ isplitMaxRightM seq root (\i x -> pure
 -- | Amortized \(O(\log n)\). Given a monotonious sequence, returns the rightmost node \(v_k\)
 -- where \(f(v)\) holds for every \([0, i) (0 \le i \lt k)\).
 --
--- ==== Constraints
--- - The node must be a root.
---
 -- @since 1.2.0.0
 {-# INLINEABLE isplitMaxRightM #-}
 isplitMaxRightM ::
@@ -961,9 +992,6 @@ isplitMaxRightM seq@Seq {..} root f
 -- | Amortized \(O(\log n)\). Given a monotonious sequence, returns the rightmost node \(v_k\)
 -- where \(f(v)\) holds for every \([0, i) (0 \le i \lt k)\).
 --
--- ==== Constraints
--- - The node must be a root.
---
 -- @since 1.2.0.0
 {-# INLINE isplitMaxRightProdST #-}
 isplitMaxRightProdST ::
@@ -981,9 +1009,6 @@ isplitMaxRightProdST seq root f = stToPrim $ isplitMaxRightProdM seq root (\i x
 -- | Amortized \(O(\log n)\). Given a monotonious sequence, returns the rightmost node \(v_k\)
 -- where \(f(v)\) holds for every \([0, i) (0 \le i \lt k)\).
 --
--- ==== Constraints
--- - The node must be a root.
---
 -- @since 1.2.0.0
 {-# INLINEABLE isplitMaxRightProdM #-}
 isplitMaxRightProdM ::
@@ -1072,7 +1097,7 @@ imaxRightM seq@Seq {..} root0 f = do
               if P.nullIndex l
                 then pure offset
                 else (offset +) <$> VGM.read sSeq (coerce l)
-            b <- f pos v
+            !b <- f pos v
             if b
               then do
                 r <- stToPrim $ VGM.read rSeq $ coerce root
@@ -1142,7 +1167,7 @@ imaxRightProdM seq@Seq {..} root0 f = do
               VGM.write rSeq (coerce root) rootR
               updateNodeST seq root
               pure $! acc <> prodRoot
-            b <- f pos prodM
+            !b <- f pos prodM
             if b
               then do
                 r <- stToPrim $ VGM.read rSeq $ coerce root
@@ -1161,30 +1186,103 @@ imaxRightProdM seq@Seq {..} root0 f = do
 -- | Amortized \(O(n)\). Returns the sequence of monoid values.
 --
 -- @since 1.2.0.0
-{-# INLINE freezeST #-}
+{-# INLINEABLE freezeST #-}
 freezeST :: (HasCallStack, SegAct f a, Eq f, Monoid f, VU.Unbox f, Monoid a, VU.Unbox a) => Seq s f a -> P.Index -> ST s (VU.Vector a)
-freezeST seq@Seq {sSeq, lSeq, rSeq, vSeq} root0 = do
-  size <- VGM.read sSeq (coerce root0)
-  res <- VUM.unsafeNew size
-  let inner i root
-        | P.nullIndex root = pure i
+freezeST seq@Seq {sSeq, lSeq, rSeq, vSeq} root0
+  | P.nullIndex root0 = pure VU.empty
+  | otherwise = do
+      size <- VGM.read sSeq (coerce root0)
+      res <- VUM.unsafeNew size
+      let inner i root
+            | P.nullIndex root = pure i
+            | otherwise = do
+                -- visit from left to right
+                propNodeST seq root
+                i' <- inner i =<< VGM.read lSeq (coerce root)
+                vx <- VGM.read vSeq (coerce root)
+                VGM.write res i' vx
+                inner (i' + 1) =<< VGM.read rSeq (coerce root)
+      _ <- inner 0 root0
+      VU.unsafeFreeze res
+
+-- -------------------------------------------------------------------------------------------------
+-- Internals and node methods
+-- -------------------------------------------------------------------------------------------------
+
+-- | Amortized \(O(\log n)\).
+--
+-- @since 1.2.1.0
+{-# INLINEABLE splitMaxRightWithST #-}
+splitMaxRightWithST ::
+  (HasCallStack, SegAct f a, Eq f, Monoid f, VU.Unbox f, Monoid a, VU.Unbox a) =>
+  -- | Sequence storage
+  Seq s f a ->
+  -- | Root node
+  P.Index ->
+  -- | User predicate \(f(i)\)
+  (P.Index -> ST s Bool) ->
+  -- | (left, right) sequences where \(f\) holds for the left
+  ST s (P.Index, P.Index)
+splitMaxRightWithST seq@Seq {..} root f
+  | P.nullIndex root = pure (P.undefIndex, P.undefIndex)
+  | otherwise = do
+      assertRootST seq root
+      (!c, !_) <- maxRightWithST seq root f
+      if P.nullIndex c
+        then do
+          -- `f` does hot hold
+          splayST seq root True
+          pure (P.undefIndex, root)
+        else do
+          splayST seq c True
+          right <- VGM.read rSeq (coerce c)
+          if P.nullIndex right
+            then do
+              -- `f` holds for the whole sequence
+              pure (c, P.undefIndex)
+            else do
+              -- `f` holds for part of the sequence. detach the right child
+              VGM.write pSeq (coerce right) P.undefIndex
+              VGM.write rSeq (coerce c) P.undefIndex
+              updateNodeST seq c
+              pure (c, right)
+
+-- | Amortized \(O(\log n)\). Given a monotonious sequence, returns the rightmost node \(v_k\)
+-- where \(f(v)\) holds for every \([0, i) (0 \le i \lt k)\).
+--
+-- ==== Constraints
+-- - The node must be a root.
+--
+-- @since 1.2.1.0
+{-# INLINEABLE maxRightWithST #-}
+maxRightWithST ::
+  (HasCallStack, SegAct f a, Eq f, Monoid f, VU.Unbox f, Monoid a, VU.Unbox a) =>
+  -- | Sequence storage
+  Seq s f a ->
+  -- | Root node
+  P.Index ->
+  -- | User predicate
+  (P.Index -> ST s Bool) ->
+  -- | (rightmost node, new root)
+  ST s (P.Index, P.Index)
+maxRightWithST seq@Seq {..} root0 f = do
+  let inner parent lastYes root
+        | P.nullIndex root = pure (lastYes, parent)
         | otherwise = do
-            -- visit from left to right
-            propNodeST seq root
-            i' <- inner i =<< VGM.read lSeq (coerce root)
-            vx <- VGM.read vSeq (coerce root)
-            VGM.write res i' vx
-            inner (i' + 1) =<< VGM.read rSeq (coerce root)
-  _ <- inner 0 root0
-  VU.unsafeFreeze res
+            stToPrim $ propNodeST seq root
+            !b <- f root
+            if b
+              then inner root root =<< VGM.read rSeq (coerce root)
+              else inner root lastYes =<< VGM.read lSeq (coerce root)
+  (!yes, !root') <- inner P.undefIndex P.undefIndex root0
+  stToPrim $ splayST seq root' True
+  pure (yes, root')
 
--- -------------------------------------------------------------------------------------------------
--- Node methods
--- -------------------------------------------------------------------------------------------------
-
--- NOTE(pref): inlining these functions are important for the speed
+-- NOTE(pref): inlining node functions are important for the speed
 
 -- | \(O(1)\) Recomputes the node size and the monoid product.
+--
+-- @since 1.2.1.0
 {-# INLINEABLE updateNodeST #-}
 updateNodeST :: (Monoid a, VU.Unbox a) => Seq s f a -> P.Index -> ST s ()
 updateNodeST Seq {..} i = do
@@ -1208,7 +1306,9 @@ updateNodeST Seq {..} i = do
   VGM.write sSeq (coerce i) size''
   VGM.write prodSeq (coerce i) prod''
 
--- | \(O(1)\) Writes to the monoid.
+-- | \(O(1)\) Writes to the monoid value of a node.
+--
+-- @since 1.2.1.0
 {-# INLINE writeNodeST #-}
 writeNodeST :: (Monoid a, VU.Unbox a) => Seq s f a -> P.Index -> a -> ST s ()
 writeNodeST seq@Seq {..} root v = do
@@ -1216,7 +1316,9 @@ writeNodeST seq@Seq {..} root v = do
   VGM.write vSeq (coerce root) v
   updateNodeST seq root
 
--- | \(O(1)\) Modifies the monoid.
+-- | \(O(1)\) Modifies the monoid value of a node.
+--
+-- @since 1.2.1.0
 {-# INLINE modifyNodeST #-}
 modifyNodeST :: (HasCallStack, Monoid a, VU.Unbox a) => Seq s f a -> (a -> a) -> P.Index -> ST s ()
 modifyNodeST seq@Seq {..} f root = do
@@ -1224,7 +1326,9 @@ modifyNodeST seq@Seq {..} f root = do
   VGM.modify vSeq f $ coerce root
   updateNodeST seq root
 
--- | \(O(1)\) Modifies the monoid.
+-- | \(O(1)\) Exchanges the monoid value of a node.
+--
+-- @since 1.2.1.0
 {-# INLINE exchangeNodeST #-}
 exchangeNodeST :: (HasCallStack, Monoid a, VU.Unbox a) => Seq s f a -> P.Index -> a -> ST s a
 exchangeNodeST seq@Seq {..} root v = do
@@ -1248,12 +1352,15 @@ reverseNodeST seq@Seq {..} i = do
   VGM.modify revSeq (xor (Bit True)) $ coerce i
 
 -- | Amortized \(O(\log n)\). Propgates the lazily propagated values on a node.
+--
+-- @since 1.2.1.0
 {-# INLINE propNodeST #-}
 -- NOTE(pref): Although this function is large, inlining it needs for the speed.
 propNodeST :: (HasCallStack, SegAct f a, Eq f, VU.Unbox f, Monoid a, VU.Unbox a) => Seq s f a -> P.Index -> ST s ()
 propNodeST seq@Seq {..} i = do
   -- action
   act <- VGM.exchange lazySeq (coerce i) mempty
+  -- this is where `Eq f` is required:
   when (act /= mempty) $ do
     l <- VGM.read lSeq $ coerce i
     unless (P.nullIndex l) $ do
@@ -1277,7 +1384,7 @@ propNodeST seq@Seq {..} i = do
 -- | Amortized \(O(\log n)\). Propagetes from the root to the given node.
 {-# INLINE propNodeFromRootST #-}
 propNodeFromRootST :: (HasCallStack, SegAct f a, VU.Unbox f, VU.Unbox a, Monoid a) => Seq s f a -> P.Index -> ST s ()
-propNodeFromRootST Seq {..} i0 = inner i0
+propNodeFromRootST Seq {..} = inner
   where
     inner i = do
       p <- VGM.read pSeq $ coerce i
@@ -1286,6 +1393,8 @@ propNodeFromRootST Seq {..} i0 = inner i0
       inner i
 
 -- | Amortized \(O(\log n)\). Propgates at a node.
+--
+-- @since 1.2.1.0
 {-# INLINE applyNodeST #-}
 applyNodeST :: (HasCallStack, SegAct f a, VU.Unbox f, Monoid a, VU.Unbox a) => Seq s f a -> P.Index -> f -> ST s ()
 applyNodeST Seq {..} i act = do
