@@ -75,6 +75,7 @@
 -- >>> --    +
 -- >>> --    +--4--5
 -- >>> let n = 6
+-- >>> -- note that the edges must be bi-directed:
 -- >>> let tree = Gr.build' n . Gr.swapDupe' $ VU.fromList [(0, 1), (1, 2), (2, 3), (1, 4), (4, 5)]
 -- >>> let hld = Hld.new tree
 --
@@ -215,107 +216,128 @@ data Hld = Hld
 
 -- | \(O(n)\) Creates an `Hld` with \(0\) as the root vertex.
 --
+-- ==== Constraints
+-- - \(n \ge 1\)
+-- - The input graph must be a tree; the edges must be non-directed (both \((u, v, w)\) and
+--   \((v, u, w)\) edges are required).
+--
 -- @since 1.1.0.0
-{-# INLINABLE new #-}
+{-# INLINEABLE new #-}
 new :: forall w. (HasCallStack) => Gr.Csr w -> Hld
 new tree = newAt tree 0
 
 -- | \(O(n)\) Creates an `Hld` with a root vertex specified.
 --
+-- ==== Constraints
+-- - \(n \ge 1\)
+-- - The input graph must be a tree; the edges must be non-directed (both \((u, v, w)\) and
+--   \((v, u, w)\) edges are required).
+--
 -- @since 1.1.0.0
-{-# INLINABLE newAt #-}
+{-# INLINEABLE newAt #-}
 newAt :: forall w. (HasCallStack) => Gr.Csr w -> Vertex -> Hld
-newAt tree root = runST $ do
-  -- Re-create adjacent vertices so that the biggest subtree's head vertex comes first.
-  --
-  -- We /could/ instead record the biggest adjacent subtree vertex for each vertex, but the other
-  -- DFS would be harder.
-  let (!tree', !parent, !depths, !subtreeSize) = runST $ do
-        adjVec <- VU.thaw (Gr.adjCsr tree)
-        parent_ <- VUM.unsafeNew n
-        depths_ <- VUM.unsafeNew n
-        subtreeSize_ <- VUM.unsafeNew n
+newAt tree root
+  | n == 1 =
+      Hld
+        0
+        (VU.singleton (-1))
+        (VU.singleton 0)
+        (VU.singleton (-1))
+        (VU.singleton 0)
+        (VU.singleton 0)
+        (VU.singleton 1)
+  | otherwise = runST $ do
+      -- Re-create adjacent vertices so that the biggest subtree's head vertex comes first.
+      --
+      -- We /could/ instead record the biggest adjacent subtree vertex for each vertex, but the other
+      -- DFS would be harder.
+      let (!tree', !parent, !depths, !subtreeSize) = runST $ do
+            adjVec <- VU.thaw (Gr.adjCsr tree)
+            parent_ <- VUM.unsafeNew n
+            depths_ <- VUM.unsafeNew n
+            subtreeSize_ <- VUM.unsafeNew n
 
-        _ <- (\f -> fix f 0 (-1) root) $ \loop depth p v1 -> do
-          VGM.write parent_ v1 p
-          VGM.write depths_ v1 depth
+            _ <- (\f -> fix f 0 (-1) root) $ \loop depth p v1 -> do
+              VGM.write parent_ v1 p
+              VGM.write depths_ v1 depth
 
-          (!size1, !eBig) <-
-            VU.foldM'
-              ( \(!size1, !eBig) (!e2, !v2) -> do
-                  if v2 == p
-                    then pure (size1, eBig)
-                    else do
-                      size2 <- loop (depth + 1) v1 v2
-                      -- NOTE: It's `>` because we should swap at least once if there's some vertex other
-                      -- that the parent_.
-                      pure (size1 + size2, if size1 > size2 then eBig else e2)
-              )
-              (1 :: Int, -1)
-              (tree `Gr.eAdj` v1)
+              (!size1, !eBig) <-
+                VU.foldM'
+                  ( \(!size1, !eBig) (!e2, !v2) -> do
+                      if v2 == p
+                        then pure (size1, eBig)
+                        else do
+                          size2 <- loop (depth + 1) v1 v2
+                          -- NOTE: It's `>` because we should swap at least once if there's some vertex other
+                          -- that the parent_.
+                          pure (size1 + size2, if size1 > size2 then eBig else e2)
+                  )
+                  (1 :: Int, -1)
+                  (tree `Gr.eAdj` v1)
 
-          -- move the biggest subtree's head to the first adjacent vertex.
-          -- it means the "heavy edge" or the longest segment.
-          when (eBig /= -1) $ do
-            VGM.swap adjVec eBig $ fst (VG.head (tree `Gr.eAdj` v1))
+              -- move the biggest subtree's head to the first adjacent vertex.
+              -- it means the "heavy edge" or the longest segment.
+              when (eBig /= -1) $ do
+                VGM.swap adjVec eBig $ fst (VG.head (tree `Gr.eAdj` v1))
 
-          -- record subtree size
-          VGM.write subtreeSize_ v1 size1
+              -- record subtree size
+              VGM.write subtreeSize_ v1 size1
 
-          pure size1
+              pure size1
 
-        !vec <- VU.unsafeFreeze adjVec
-        (tree {Gr.adjCsr = vec},,,)
-          <$> VU.unsafeFreeze parent_
-          <*> VU.unsafeFreeze depths_
-          <*> VU.unsafeFreeze subtreeSize_
+            !vec <- VU.unsafeFreeze adjVec
+            (tree {Gr.adjCsr = vec},,,)
+              <$> VU.unsafeFreeze parent_
+              <*> VU.unsafeFreeze depths_
+              <*> VU.unsafeFreeze subtreeSize_
 
-  -- vertex -> reindexed vertex index
-  indices <- VUM.replicate n (-1 :: Int)
+      -- vertex -> reindexed vertex index
+      indices <- VUM.replicate n (-1 :: Int)
 
-  -- vertex -> head vertex of the segment
-  heads <- VUM.replicate n (-1 :: Int)
+      -- vertex -> head vertex of the segment
+      heads <- VUM.replicate n (-1 :: Int)
 
-  _ <- (\f -> fix f (0 :: Int) root (-1) root) $ \loop acc h p v1 -> do
-    -- reindex:
-    VGM.write indices v1 acc
-    let !acc' = acc + 1
+      _ <- (\f -> fix f (0 :: Int) root (-1) root) $ \loop acc h p v1 -> do
+        -- reindex:
+        VGM.write indices v1 acc
+        let !acc' = acc + 1
 
-    VGM.write heads v1 h
+        VGM.write heads v1 h
 
-    -- when the first vertex is within the same segment:
-    let (!adj1, !rest) = fromJust $ VU.uncons (tree' `Gr.adj` v1)
-    acc'' <-
-      if adj1 == p
-        then pure acc'
-        else loop acc' h v1 adj1
+        -- when the first vertex is within the same segment:
+        let (!adj1, !rest) = fromJust $ VU.uncons (tree' `Gr.adj` v1)
+        acc'' <-
+          if adj1 == p
+            then pure acc'
+            else loop acc' h v1 adj1
 
-    -- the others are in other segments:
-    VU.foldM'
-      ( \a v2 -> do
-          if v2 == p
-            then pure a
-            else loop a v2 v1 v2
-      )
-      acc''
-      rest
+        -- the others are in other segments:
+        VU.foldM'
+          ( \a v2 -> do
+              if v2 == p
+                then pure a
+                else loop a v2 v1 v2
+          )
+          acc''
+          rest
 
-  !indices' <- VU.unsafeFreeze indices
-  let !revIndex = VU.update (VU.replicate n (-1)) $ VU.imap (flip (,)) indices'
+      !indices' <- VU.unsafeFreeze indices
+      let !revIndex = VU.update (VU.replicate n (-1)) $ VU.imap (flip (,)) indices'
 
-  Hld root parent indices'
-    <$> VU.unsafeFreeze heads
-    <*> pure revIndex
-    <*> pure depths
-    <*> pure subtreeSize
+      Hld root parent indices'
+        <$> VU.unsafeFreeze heads
+        <*> pure revIndex
+        <*> pure depths
+        <*> pure subtreeSize
   where
     !n = Gr.nCsr tree
     !_ = ACIA.runtimeAssert (2 * (Gr.nCsr tree - 1) == Gr.mCsr tree) "AtCoder.Extra.Hld.newAt: not a non-directed tree"
+    !_ = ACIA.runtimeAssert (n >= 1) "AtCoder.Extra.Hld.newAt: the tree must have at least one vertex"
 
 -- | \(O(\log n)\) Calculates the lowest common ancestor of \(u\) and \(v\).
 --
 -- @since 1.1.0.0
-{-# INLINABLE lca #-}
+{-# INLINEABLE lca #-}
 lca :: (HasCallStack) => Hld -> Vertex -> Vertex -> Vertex
 lca Hld {..} = inner
   where
@@ -338,7 +360,7 @@ lca Hld {..} = inner
 -- is bigger than the depth of \(v\).
 --
 -- @since 1.1.0.0
-{-# INLINABLE ancestor #-}
+{-# INLINEABLE ancestor #-}
 ancestor :: (HasCallStack) => Hld -> Vertex -> Int -> Vertex
 ancestor Hld {..} parent k0 = inner parent k0
   where
@@ -357,7 +379,7 @@ ancestor Hld {..} parent k0 = inner parent k0
 -- Throws an error if `k` is out
 --
 -- @since 1.1.0.0
-{-# INLINABLE jump #-}
+{-# INLINEABLE jump #-}
 jump :: (HasCallStack) => Hld -> Vertex -> Vertex -> Int -> Maybe Vertex
 jump hld@Hld {..} u v k
   | k > lenU + lenV = Nothing
@@ -373,7 +395,7 @@ jump hld@Hld {..} u v k
 -- | \(O(\log n)\) Returns the length of the path between \(u\) and \(v\).
 --
 -- @since 1.1.0.0
-{-# INLINABLE lengthBetween #-}
+{-# INLINEABLE lengthBetween #-}
 lengthBetween :: (HasCallStack) => Hld -> Vertex -> Vertex -> Int
 lengthBetween hld@Hld {..} u v = du - dLca + dv - dLca
   where
@@ -385,7 +407,7 @@ lengthBetween hld@Hld {..} u v = du - dLca + dv - dLca
 -- | \(O(n)\) Returns the vertices on the path between \(u\) and \(v\).
 --
 -- @since 1.1.0.0
-{-# INLINABLE path #-}
+{-# INLINEABLE path #-}
 path :: (HasCallStack) => Hld -> Vertex -> Vertex -> [Vertex]
 path hld@Hld {..} u v = concatMap expand $ pathSegmentsInclusive WeightsAreOnVertices hld u v
   where
@@ -400,7 +422,7 @@ path hld@Hld {..} u v = concatMap expand $ pathSegmentsInclusive WeightsAreOnVer
 -- `WeightsAreOnEdges`. This is the trick to put edge weights to on vertices.
 --
 -- @since 1.1.0.0
-{-# INLINABLE pathSegmentsInclusive #-}
+{-# INLINEABLE pathSegmentsInclusive #-}
 pathSegmentsInclusive :: (HasCallStack) => WeightPolicy -> Hld -> Vertex -> Vertex -> [(VertexHld, VertexHld)]
 pathSegmentsInclusive weightPolicy Hld {..} x0 y0 = done $ inner x0 [] y0 []
   where
@@ -440,7 +462,7 @@ pathSegmentsInclusive weightPolicy Hld {..} x0 y0 = done $ inner x0 [] y0 []
 -- corresponds to the subtree segments rooted at the given @subtreeRoot@.
 --
 -- @since 1.1.0.0
-{-# INLINABLE subtreeSegmentInclusive #-}
+{-# INLINEABLE subtreeSegmentInclusive #-}
 subtreeSegmentInclusive :: (HasCallStack) => Hld -> Vertex -> (VertexHld, VertexHld)
 subtreeSegmentInclusive Hld {..} subtreeRoot = (ir, ir + sr - 1)
   where
@@ -450,7 +472,7 @@ subtreeSegmentInclusive Hld {..} subtreeRoot = (ir, ir + sr - 1)
 -- | \(O(1)\) Returns `True` if \(u\) is in a subtree of \(r\).
 --
 -- @since 1.1.0.0
-{-# INLINABLE isInSubtree #-}
+{-# INLINEABLE isInSubtree #-}
 isInSubtree :: (HasCallStack) => Hld -> Vertex -> Vertex -> Bool
 isInSubtree hld@Hld {..} r_ u = l <= iu && iu <= r
   where
