@@ -79,7 +79,7 @@
 --
 -- @since 1.0.0.0
 module AtCoder.LazySegTree
-  ( -- Lazy segment tree
+  ( -- * Lazy segment tree
     SegAct (..),
     LazySegTree (nLst, sizeLst, logLst),
 
@@ -97,7 +97,7 @@ module AtCoder.LazySegTree
     -- * Products
     prod,
     prodMaybe,
-    allProd,
+    allProd, -- FIXME: rename to prodAll
 
     -- * Applications
     applyAt,
@@ -143,7 +143,7 @@ import Prelude hiding (read)
 --
 -- If you implement `SegAct` via `segActWithLength`, satisfy one more propety:
 --
--- [Linear left monoid action] @'segActWithLength' len f a = 'Data.Semigroup.stimes' len ('segAct' f a) a@.
+-- [Linear left monoid action] @'segActWithLength' len f (stimes len a) = 'Data.Semigroup.stimes' len ('segAct' f a)@.
 --
 -- ==== Invariant
 -- In `SegAct` instances, new semigroup values are always given from the left: @new '<>' old@. The
@@ -198,6 +198,7 @@ import Prelude hiding (read)
 -- Define your monoid action @F@ and your acted monoid @X@:
 --
 -- @
+-- {-# LANGUAGE DerivingStrategies #-}
 -- {-# LANGUAGE TypeFamilies #-}
 --
 -- import AtCoder.LazySegTree qualified as LST
@@ -320,21 +321,6 @@ data LazySegTree s f a = LazySegTree
     lzLst :: !(VUM.MVector s f)
   }
 
-{-# INLINE buildST #-}
-buildST :: (Monoid f, VU.Unbox f, Monoid a, VU.Unbox a) => VU.Vector a -> ST s (LazySegTree s f a)
-buildST vs = do
-  let nLst = VU.length vs
-  let sizeLst = ACIBIT.bitCeil nLst
-  let logLst = countTrailingZeros sizeLst
-  dLst <- VUM.replicate (2 * sizeLst) mempty
-  lzLst <- VUM.replicate sizeLst mempty
-  VU.iforM_ vs $ \i v -> do
-    VGM.write dLst (sizeLst + i) v
-  let segtree = LazySegTree {..}
-  for_ [sizeLst - 1, sizeLst - 2 .. 1] $ \i -> do
-    updateST segtree i
-  pure segtree
-
 -- | Creates an array of length \(n\). All the elements are initialized to `mempty`.
 --
 -- ==== Constraints
@@ -363,16 +349,6 @@ new nLst
 build :: (PrimMonad m, Monoid f, VU.Unbox f, Monoid a, VU.Unbox a) => VU.Vector a -> m (LazySegTree (PrimState m) f a)
 build vs = stToPrim $ buildST vs
 
-{-# INLINE writeST #-}
-writeST :: (SegAct f a, VU.Unbox f, Monoid a, VU.Unbox a) => LazySegTree s f a -> Int -> a -> ST s ()
-writeST self@LazySegTree {..} p x = do
-  let p' = p + sizeLst
-  for_ [logLst, logLst - 1 .. 1] $ \i -> do
-    pushST self $ p' .>>. i
-  VGM.unsafeWrite dLst p' x
-  for_ [1 .. logLst] $ \i -> do
-    updateST self $ p' .>>. i
-
 -- | Sets \(p\)-th value of the array to \(x\).
 --
 -- ==== Constraints
@@ -388,16 +364,6 @@ write self p x = stToPrim $ writeST self p x
   where
     !_ = ACIA.checkIndex "AtCoder.LazySegTree.write" p (nLst self)
 
-{-# INLINE modifyST #-}
-modifyST :: (SegAct f a, VU.Unbox f, Monoid a, VU.Unbox a) => LazySegTree s f a -> (a -> a) -> Int -> ST s ()
-modifyST self@LazySegTree {..} f p = do
-  let p' = p + sizeLst
-  for_ [logLst, logLst - 1 .. 1] $ \i -> do
-    pushST self $ p' .>>. i
-  VGM.unsafeModify dLst f p'
-  for_ [1 .. logLst] $ \i -> do
-    updateST self $ p' .>>. i
-
 -- | (Extra API) Modifies \(p\)-th value with a function \(f\).
 --
 -- ==== Constraints
@@ -422,7 +388,7 @@ modify self f p = stToPrim $ modifyST self f p
 -- - \(O(\log n)\)
 --
 -- @since 1.0.0.0
-{-# INLINE modifyM #-}
+{-# INLINEABLE modifyM #-}
 modifyM :: (HasCallStack, PrimMonad m, SegAct f a, VU.Unbox f, Monoid a, VU.Unbox a) => LazySegTree (PrimState m) f a -> (a -> m a) -> Int -> m ()
 modifyM self@LazySegTree {..} f p = do
   let !_ = ACIA.checkIndex "AtCoder.LazySegTree.modifyM" p nLst
@@ -433,17 +399,6 @@ modifyM self@LazySegTree {..} f p = do
   stToPrim $ for_ [1 .. logLst] $ \i -> do
     updateST self $ p' .>>. i
 
-{-# INLINE exchangeST #-}
-exchangeST :: (SegAct f a, VU.Unbox f, Monoid a, VU.Unbox a) => LazySegTree s f a -> Int -> a -> ST s a
-exchangeST self@LazySegTree {..} p x = do
-  let p' = p + sizeLst
-  for_ [logLst, logLst - 1 .. 1] $ \i -> do
-    pushST self $ p' .>>. i
-  res <- VGM.unsafeExchange dLst p' x
-  for_ [1 .. logLst] $ \i -> do
-    updateST self $ p' .>>. i
-  pure res
-
 -- | (Extra API) Sets \(p\)-th value of the array to \(x\) and returns the old value.
 --
 -- ==== Constraints
@@ -459,14 +414,6 @@ exchange self p x = stToPrim $ exchangeST self p x
   where
     !_ = ACIA.checkIndex "AtCoder.LazySegTree.exchange" p (nLst self)
 
-{-# INLINE readST #-}
-readST :: (SegAct f a, VU.Unbox f, Monoid a, VU.Unbox a) => LazySegTree s f a -> Int -> ST s a
-readST self@LazySegTree {..} p = do
-  let p' = p + sizeLst
-  for_ [logLst, logLst - 1 .. 1] $ \i -> do
-    pushST self $ p' .>>. i
-  VGM.unsafeRead dLst p'
-
 -- | Returns \(p\)-th value of the array.
 --
 -- ==== Constraints
@@ -514,31 +461,6 @@ prodMaybe self@LazySegTree {nLst} l0 r0
   | l0 == r0 = pure (Just mempty)
   | otherwise = stToPrim $ Just <$> unsafeProdST self l0 r0
 
--- | Internal implementation of `prod`.
-{-# INLINE unsafeProdST #-}
-unsafeProdST :: (SegAct f a, VU.Unbox f, Monoid a, VU.Unbox a) => LazySegTree s f a -> Int -> Int -> ST s a
-unsafeProdST self@LazySegTree {..} l0 r0 = do
-  let l = l0 + sizeLst
-  let r = r0 + sizeLst
-  for_ [logLst, logLst - 1 .. 1] $ \i -> do
-    when (((l .>>. i) .<<. i) /= l) $ pushST self $ l .>>. i
-    when (((r .>>. i) .<<. i) /= r) $ pushST self $ (r - 1) .>>. i
-  inner l (r - 1) mempty mempty
-  where
-    -- NOTE: we're using inclusive range [l, r] for simplicity
-    inner l r !smL !smR
-      | l > r = pure $! smL <> smR
-      | otherwise = do
-          !smL' <-
-            if testBit l 0
-              then (smL <>) <$> VGM.read dLst l
-              else pure smL
-          !smR' <-
-            if not $ testBit r 0
-              then (<> smR) <$> VGM.read dLst r
-              else pure smR
-          inner ((l + 1) .>>. 1) ((r - 1) .>>. 1) smL' smR'
-
 -- | Returns the product of \([op(a[0], ..., a[n - 1])]\), assuming the properties of the monoid. It
 -- returns `mempty` if \(n = 0\).
 --
@@ -550,19 +472,6 @@ unsafeProdST self@LazySegTree {..} l0 r0 = do
 allProd :: (PrimMonad m, Monoid a, VU.Unbox a) => LazySegTree (PrimState m) f a -> m a
 allProd LazySegTree {..} = VGM.read dLst 1
 
-{-# INLINE applyAtST #-}
-applyAtST :: (SegAct f a, VU.Unbox f, Monoid a, VU.Unbox a) => LazySegTree s f a -> Int -> f -> ST s ()
-applyAtST self@LazySegTree {..} p f = do
-  let p' = p + sizeLst
-  -- propagate
-  for_ [logLst, logLst - 1 .. 1] $ \i -> do
-    pushST self $ p' .>>. i
-  let !len = bit $! logLst - (63 - countLeadingZeros p')
-  VGM.modify dLst (segActWithLength len f) p'
-  -- evaluate
-  for_ [1 .. logLst] $ \i -> do
-    updateST self $ p' .>>. i
-
 -- | Applies @segAct f@ to an index \(p\).
 --
 -- ==== Constraints
@@ -578,33 +487,6 @@ applyAt self p f = stToPrim $ applyAtST self p f
   where
     !_ = ACIA.checkIndex "AtCoder.LazySegTree.applyAt" p (nLst self)
 
-{-# INLINE applyInST #-}
-applyInST :: (SegAct f a, VU.Unbox f, Monoid a, VU.Unbox a) => LazySegTree s f a -> Int -> Int -> f -> ST s ()
-applyInST self@LazySegTree {..} l0 r0 f
-  | l0 == r0 = pure ()
-  | otherwise = do
-      let l = l0 + sizeLst
-      let r = r0 + sizeLst
-      -- propagate
-      for_ [logLst, logLst - 1 .. 1] $ \i -> do
-        when (((l .>>. i) .<<. i) /= l) $ pushST self (l .>>. i)
-        when (((r .>>. i) .<<. i) /= r) $ pushST self ((r - 1) .>>. i)
-      inner l (r - 1)
-      -- evaluate
-      for_ [1 .. logLst] $ \i -> do
-        when (((l .>>. i) .<<. i) /= l) $ updateST self (l .>>. i)
-        when (((r .>>. i) .<<. i) /= r) $ updateST self ((r - 1) .>>. i)
-  where
-    -- NOTE: we're using inclusive range [l, r] for simplicity
-    inner l r
-      | l > r = pure ()
-      | otherwise = do
-          when (testBit l 0) $ do
-            allApplyST self l f
-          unless (testBit r 0) $ do
-            allApplyST self r f
-          inner ((l + 1) .>>. 1) ((r - 1) .>>. 1)
-
 -- | Applies @segAct f@ to an interval \([l, r)\).
 --
 -- ==== Constraints
@@ -655,7 +537,7 @@ minLeft seg r0 g = minLeftM seg r0 (pure . g)
 -- - \(O(\log n)\)
 --
 -- @since 1.0.0.0
-{-# INLINE minLeftM #-}
+{-# INLINEABLE minLeftM #-}
 minLeftM :: (HasCallStack, PrimMonad m, SegAct f a, VU.Unbox f, Monoid a, VU.Unbox a) => LazySegTree (PrimState m) f a -> Int -> (a -> m Bool) -> m Int
 minLeftM self@LazySegTree {..} r0 g = do
   b <- g mempty
@@ -731,11 +613,11 @@ maxRight seg l0 g = maxRightM seg l0 (pure . g)
 -- - \(O(\log n)\)
 --
 -- @since 1.0.0.0
-{-# INLINE maxRightM #-}
+{-# INLINEABLE maxRightM #-}
 maxRightM :: (HasCallStack, PrimMonad m, SegAct f a, VU.Unbox f, Monoid a, VU.Unbox a) => LazySegTree (PrimState m) f a -> Int -> (a -> m Bool) -> m Int
 maxRightM self@LazySegTree {..} l0 g = do
   b <- g mempty
-  let !_ = ACIA.runtimeAssert b "AtCoder.LazySegTree.maxRightM: `g mempty` returned `False`"
+  let !_ = ACIA.runtimeAssert b "AtCoder.LazySegTree.maxRightM: `g mempty` must return `False`"
   if l0 == nLst
     then pure nLst
     else do
@@ -748,9 +630,9 @@ maxRightM self@LazySegTree {..} l0 g = do
     !_ = ACIA.runtimeAssert (0 <= l0 && l0 <= nLst) $ "AtCoder.LazySegTree.maxRightM: given invalid `left` index `" ++ show l0 ++ "` over length `" ++ show nLst ++ "`"
     inner l !sm = do
       let l' = chooseBit l
-      !sm' <- (sm <>) <$> VGM.read dLst l'
+      !sm' <- stToPrim $ (sm <>) <$> VGM.read dLst l'
       b <- g sm'
-      if not $ b
+      if not b
         then do
           inner2 l' sm
         else do
@@ -779,11 +661,7 @@ maxRightM self@LazySegTree {..} l0 g = do
 -- @since 1.0.0.0
 {-# INLINE freeze #-}
 freeze :: (PrimMonad m, SegAct f a, VU.Unbox f, Monoid a, VU.Unbox a) => LazySegTree (PrimState m) f a -> m (VU.Vector a)
-freeze self@LazySegTree {..} = do
-  -- push all (we _could_ skip some elements)
-  stToPrim $ for_ [1 .. sizeLst - 1] $ \i -> do
-    pushST self i
-  VU.freeze . VUM.take nLst $ VUM.drop sizeLst dLst
+freeze = stToPrim . freezeST
 
 -- | \(O(n)\) Unsafely converts a mutable vector to an immutable one without copying. The mutable
 -- vector may not be used after this operation.
@@ -791,14 +669,149 @@ freeze self@LazySegTree {..} = do
 -- @since 1.0.0.0
 {-# INLINE unsafeFreeze #-}
 unsafeFreeze :: (PrimMonad m, SegAct f a, VU.Unbox f, Monoid a, VU.Unbox a) => LazySegTree (PrimState m) f a -> m (VU.Vector a)
-unsafeFreeze self@LazySegTree {..} = do
-  -- push all (we _could_ skip some elements)
-  stToPrim $ for_ [1 .. sizeLst - 1] $ \i -> do
-    pushST self i
-  VU.unsafeFreeze . VUM.take nLst $ VUM.drop sizeLst dLst
+unsafeFreeze = stToPrim . unsafeFreezeST
 
 -- NOTE (perf): these functions have to be inlined after all:
 
+-- -------------------------------------------------------------------------------------------------
+-- Internal
+-- -------------------------------------------------------------------------------------------------
+
+{-# INLINEABLE buildST #-}
+buildST :: (Monoid f, VU.Unbox f, Monoid a, VU.Unbox a) => VU.Vector a -> ST s (LazySegTree s f a)
+buildST vs = do
+  let nLst = VU.length vs
+  let sizeLst = ACIBIT.bitCeil nLst
+  let logLst = countTrailingZeros sizeLst
+  dLst <- VUM.replicate (2 * sizeLst) mempty
+  lzLst <- VUM.replicate sizeLst mempty
+  VU.iforM_ vs $ \i v -> do
+    VGM.write dLst (sizeLst + i) v
+  let segtree = LazySegTree {..}
+  for_ [sizeLst - 1, sizeLst - 2 .. 1] $ \i -> do
+    updateST segtree i
+  pure segtree
+
+{-# INLINEABLE writeST #-}
+writeST :: (SegAct f a, VU.Unbox f, Monoid a, VU.Unbox a) => LazySegTree s f a -> Int -> a -> ST s ()
+writeST self@LazySegTree {..} p x = do
+  let p' = p + sizeLst
+  for_ [logLst, logLst - 1 .. 1] $ \i -> do
+    pushST self $ p' .>>. i
+  VGM.unsafeWrite dLst p' x
+  for_ [1 .. logLst] $ \i -> do
+    updateST self $ p' .>>. i
+
+{-# INLINEABLE modifyST #-}
+modifyST :: (SegAct f a, VU.Unbox f, Monoid a, VU.Unbox a) => LazySegTree s f a -> (a -> a) -> Int -> ST s ()
+modifyST self@LazySegTree {..} f p = do
+  let p' = p + sizeLst
+  for_ [logLst, logLst - 1 .. 1] $ \i -> do
+    pushST self $ p' .>>. i
+  VGM.unsafeModify dLst f p'
+  for_ [1 .. logLst] $ \i -> do
+    updateST self $ p' .>>. i
+
+{-# INLINEABLE exchangeST #-}
+exchangeST :: (SegAct f a, VU.Unbox f, Monoid a, VU.Unbox a) => LazySegTree s f a -> Int -> a -> ST s a
+exchangeST self@LazySegTree {..} p x = do
+  let p' = p + sizeLst
+  for_ [logLst, logLst - 1 .. 1] $ \i -> do
+    pushST self $ p' .>>. i
+  res <- VGM.unsafeExchange dLst p' x
+  for_ [1 .. logLst] $ \i -> do
+    updateST self $ p' .>>. i
+  pure res
+
+{-# INLINEABLE readST #-}
+readST :: (SegAct f a, VU.Unbox f, Monoid a, VU.Unbox a) => LazySegTree s f a -> Int -> ST s a
+readST self@LazySegTree {..} p = do
+  let p' = p + sizeLst
+  for_ [logLst, logLst - 1 .. 1] $ \i -> do
+    pushST self $ p' .>>. i
+  VGM.unsafeRead dLst p'
+
+{-# INLINEABLE applyAtST #-}
+applyAtST :: (SegAct f a, VU.Unbox f, Monoid a, VU.Unbox a) => LazySegTree s f a -> Int -> f -> ST s ()
+applyAtST self@LazySegTree {..} p f = do
+  let p' = p + sizeLst
+  -- propagate
+  for_ [logLst, logLst - 1 .. 1] $ \i -> do
+    pushST self $ p' .>>. i
+  -- FIXME: length should be always `1`
+  let !len = bit $! logLst - (63 - countLeadingZeros p')
+  VGM.modify dLst (segActWithLength len f) p'
+  -- evaluate
+  for_ [1 .. logLst] $ \i -> do
+    updateST self $ p' .>>. i
+
+{-# INLINEABLE applyInST #-}
+applyInST :: (SegAct f a, VU.Unbox f, Monoid a, VU.Unbox a) => LazySegTree s f a -> Int -> Int -> f -> ST s ()
+applyInST self@LazySegTree {..} l0 r0 f
+  | l0 == r0 = pure ()
+  | otherwise = do
+      let l = l0 + sizeLst
+      let r = r0 + sizeLst
+      -- propagate
+      for_ [logLst, logLst - 1 .. 1] $ \i -> do
+        when (((l .>>. i) .<<. i) /= l) $ pushST self (l .>>. i)
+        when (((r .>>. i) .<<. i) /= r) $ pushST self ((r - 1) .>>. i)
+      inner l (r - 1)
+      -- evaluate
+      for_ [1 .. logLst] $ \i -> do
+        when (((l .>>. i) .<<. i) /= l) $ updateST self (l .>>. i)
+        when (((r .>>. i) .<<. i) /= r) $ updateST self ((r - 1) .>>. i)
+  where
+    -- NOTE: we're using inclusive range [l, r] for simplicity
+    inner l r
+      | l > r = pure ()
+      | otherwise = do
+          when (testBit l 0) $ do
+            allApplyST self l f
+          unless (testBit r 0) $ do
+            allApplyST self r f
+          inner ((l + 1) .>>. 1) ((r - 1) .>>. 1)
+
+-- | Internal implementation of `prod`.
+{-# INLINEABLE unsafeProdST #-}
+unsafeProdST :: (SegAct f a, VU.Unbox f, Monoid a, VU.Unbox a) => LazySegTree s f a -> Int -> Int -> ST s a
+unsafeProdST self@LazySegTree {..} l0 r0 = do
+  let l = l0 + sizeLst
+  let r = r0 + sizeLst
+  for_ [logLst, logLst - 1 .. 1] $ \i -> do
+    when (((l .>>. i) .<<. i) /= l) $ pushST self $ l .>>. i
+    when (((r .>>. i) .<<. i) /= r) $ pushST self $ (r - 1) .>>. i
+  inner l (r - 1) mempty mempty
+  where
+    -- NOTE: we're using inclusive range [l, r] for simplicity
+    inner l r !smL !smR
+      | l > r = pure $! smL <> smR
+      | otherwise = do
+          !smL' <-
+            if testBit l 0
+              then (smL <>) <$> VGM.read dLst l
+              else pure smL
+          !smR' <-
+            if not $ testBit r 0
+              then (<> smR) <$> VGM.read dLst r
+              else pure smR
+          inner ((l + 1) .>>. 1) ((r - 1) .>>. 1) smL' smR'
+
+{-# INLINEABLE freezeST #-}
+freezeST :: (SegAct f a, VU.Unbox f, Monoid a, VU.Unbox a) => LazySegTree s f a -> ST s (VU.Vector a)
+freezeST self@LazySegTree {..} = do
+  for_ [1 .. sizeLst - 1] $ \i -> do
+    pushST self i
+  VU.freeze . VUM.take nLst $ VUM.drop sizeLst dLst
+
+{-# INLINEABLE unsafeFreezeST #-}
+unsafeFreezeST :: (SegAct f a, VU.Unbox f, Monoid a, VU.Unbox a) => LazySegTree s f a -> ST s (VU.Vector a)
+unsafeFreezeST self@LazySegTree {..} = do
+  -- push all (we _could_ skip some elements)
+  for_ [1 .. sizeLst - 1] $ \i -> do
+    pushST self i
+  VU.unsafeFreeze . VUM.take nLst $ VUM.drop sizeLst dLst
+
 -- | \(O(1)\)
 {-# INLINE updateST #-}
 updateST :: (Monoid f, VU.Unbox f, Monoid a, VU.Unbox a) => LazySegTree s f a -> Int -> ST s ()
