@@ -5,16 +5,27 @@ import Test.Tasty.Ingredients.Rerun
 import Tests.Convolution qualified
 import Tests.Dsu qualified
 import Tests.Extra.Bisect qualified
+import Tests.Extra.DynLazySegTree qualified
+import Tests.Extra.DynLazySegTree.Persistent qualified
+import Tests.Extra.DynSegTree qualified
+import Tests.Extra.DynSegTree.Persistent qualified
+import Tests.Extra.DynSparseSegTree qualified
+import Tests.Extra.DynSparseSegTree.Persistent qualified
 import Tests.Extra.HashMap qualified
 import Tests.Extra.IntMap qualified
 import Tests.Extra.IntSet qualified
 import Tests.Extra.IntervalMap qualified
+import Tests.Extra.KdTree qualified
+import Tests.Extra.LazyKdTree qualified
 import Tests.Extra.Math qualified
 import Tests.Extra.Monoid qualified
 import Tests.Extra.MultiSet qualified
+import Tests.Extra.SegTree2d qualified
+import Tests.Extra.SegTree2d.Dense qualified
 import Tests.Extra.Semigroup.Matrix qualified
 import Tests.Extra.Semigroup.Permutation qualified
 import Tests.Extra.Seq qualified
+import Tests.Extra.Seq.Map qualified
 import Tests.Extra.WaveletMatrix qualified
 import Tests.Extra.WaveletMatrix.BitVector qualified
 import Tests.Extra.WaveletMatrix.Raw qualified
@@ -47,16 +58,27 @@ main =
         testGroup
           "Extra"
           [ testGroup "Bisect" Tests.Extra.Bisect.tests,
+            testGroup "DynLazySegTree" Tests.Extra.DynLazySegTree.tests,
+            testGroup "DynLazySegTree.Persistent" Tests.Extra.DynLazySegTree.Persistent.tests,
+            testGroup "DynSegTree" Tests.Extra.DynSegTree.tests,
+            testGroup "DynSegTree.Persistent" Tests.Extra.DynSegTree.Persistent.tests,
+            testGroup "DynSparseSegTree" Tests.Extra.DynSparseSegTree.tests,
+            testGroup "DynSparseSegTree.Persistent" Tests.Extra.DynSparseSegTree.Persistent.tests,
             testGroup "HashMap" Tests.Extra.HashMap.tests,
             testGroup "IntervalMap" Tests.Extra.IntervalMap.tests,
             testGroup "IntMap" Tests.Extra.IntMap.tests,
             testGroup "IntSet" Tests.Extra.IntSet.tests,
+            testGroup "KdTree" Tests.Extra.KdTree.tests,
+            testGroup "LazyKdTree" Tests.Extra.LazyKdTree.tests,
             testGroup "Math" Tests.Extra.Math.tests,
             testGroup "Monoid" Tests.Extra.Monoid.tests,
             testGroup "MultiSet" Tests.Extra.MultiSet.tests,
+            testGroup "SegTree2d" Tests.Extra.SegTree2d.tests,
+            testGroup "SegTree2d.Dense" Tests.Extra.SegTree2d.Dense.tests,
             testGroup "Semigroup.Matrix" Tests.Extra.Semigroup.Matrix.tests,
             testGroup "Semigroup.Permutation" Tests.Extra.Semigroup.Permutation.tests,
             testGroup "Seq" Tests.Extra.Seq.tests,
+            testGroup "Seq.Map" Tests.Extra.Seq.Map.tests,
             testGroup "WaveletMatrix" Tests.Extra.WaveletMatrix.tests,
             testGroup "WaveletMatrix.BitVector" Tests.Extra.WaveletMatrix.BitVector.tests,
             testGroup "WaveletMatrix.Raw" Tests.Extra.WaveletMatrix.Raw.tests,
